use std::{borrow::Cow, marker::PhantomData};

pub use fdb_schema_macros::schema;

pub trait ByteRepresentation {
    type Value: Sized;
    type DeserializationError: std::error::Error;
    type SerializationError: std::error::Error;
    fn from_bytes(bytes: &[u8]) -> Result<Self::Value, Self::DeserializationError>;
    fn to_bytes(value: &Self::Value) -> Result<Cow<[u8]>, Self::SerializationError>;
}

pub trait FdbTypedKey {
    type Representation: ByteRepresentation;
    fn get_raw_key(&self) -> &[u8];
}

#[derive(Debug, thiserror::Error)]
pub enum Error<T: std::error::Error> {
    #[error("Foundation DB error: {0}")]
    FoundationDb(foundationdb::FdbError),
    #[error("Serialization error: {0}")]
    Deserialization(T),
}

pub trait FdbTransactionExt {
    fn set_value<Key: FdbTypedKey>(
        &self,
        key: &Key,
        value: &<Key::Representation as ByteRepresentation>::Value,
    ) -> Result<(), <Key::Representation as ByteRepresentation>::SerializationError>;

    fn get_value<Key: FdbTypedKey>(
        &self,
        key: &Key,
    ) -> impl Future<
        Output = Result<Option<<Key::Representation as ByteRepresentation>::Value>, Error<<Key::Representation as ByteRepresentation>::DeserializationError>>,
    >;

    fn get_snapshot<Key: FdbTypedKey>(
        &self,
        key: &Key,
    ) -> impl Future<
        Output = Result<Option<<Key::Representation as ByteRepresentation>::Value>, Error<<Key::Representation as ByteRepresentation>::DeserializationError>>,
    >;
}
impl <T: std::error::Error + Send + Sync + 'static> From<Error<T>> for foundationdb::FdbBindingError {
    fn from(value: Error<T>) -> Self {
        match value {
            Error::FoundationDb(error) => error.into(),
            Error::Deserialization(error) => Self::new_custom_error(Box::new(error)),
        }
    }
}
impl FdbTransactionExt for foundationdb::Transaction {
    fn set_value<Key: FdbTypedKey>(
        &self,
        key: &Key,
        value: &<Key::Representation as ByteRepresentation>::Value,
    ) -> Result<(), <Key::Representation as ByteRepresentation>::SerializationError> {
        Key::Representation::to_bytes(value).map(|bytes| self.set(key.get_raw_key(), &bytes))
    }

    async fn get_value<Key: FdbTypedKey>(
        &self,
        key: &Key,
    ) -> Result<Option<<Key::Representation as ByteRepresentation>::Value>, Error<<Key::Representation as ByteRepresentation>::DeserializationError>> {
        self.get(key.get_raw_key(), false)
            .await
            .map_err(Error::FoundationDb)
            .and_then(|maybe| {
                maybe
                    .map(|b| Key::Representation::from_bytes(&b))
                    .transpose()
                    .map_err(Error::Deserialization)
            })
    }

    async fn get_snapshot<Key: FdbTypedKey>(
        &self,
        key: &Key,
    ) -> Result<Option<<Key::Representation as ByteRepresentation>::Value>, Error<<Key::Representation as ByteRepresentation>::DeserializationError>> {
        self.get(key.get_raw_key(), true)
            .await
            .map_err(Error::FoundationDb)
            .and_then(|maybe| {
                maybe
                    .map(|b| Key::Representation::from_bytes(&b))
                    .transpose()
                    .map_err(Error::Deserialization)
            })
    }
}

#[cfg(feature = "serde_json")]
pub struct Json<T: serde::de::DeserializeOwned + serde::Serialize>(PhantomData<T>);

#[cfg(feature = "serde_json")]
impl<T: serde::de::DeserializeOwned + serde::Serialize> ByteRepresentation for Json<T> {
    type Value = T;
    type DeserializationError = serde_json::Error;
    type SerializationError = serde_json::Error;

    fn from_bytes(bytes: &[u8]) -> Result<Self::Value, Self::DeserializationError> {
        serde_json::from_slice(bytes)
    }

    fn to_bytes(value: &Self::Value) -> Result<Cow<[u8]>, Self::SerializationError> {
        serde_json::to_vec(value).map(Cow::Owned)
    }
}

pub struct Tuple<T: foundationdb::tuple::TuplePack + for<'a> foundationdb::tuple::TupleUnpack<'a>>(
    PhantomData<T>,
);

impl<T: foundationdb::tuple::TuplePack> ByteRepresentation for Tuple<T>
where
    for<'a> T: foundationdb::tuple::TupleUnpack<'a>,
{
    type Value = T;
    type DeserializationError = foundationdb::tuple::PackError;
    type SerializationError = std::convert::Infallible;

    fn from_bytes(bytes: &[u8]) -> Result<Self::Value, Self::DeserializationError> {
        foundationdb::tuple::unpack(bytes)
    }

    fn to_bytes(value: &Self::Value) -> Result<Cow<[u8]>, Self::SerializationError> {
        Ok(Cow::Owned(foundationdb::tuple::pack(value)))
    }
}

pub struct Raw<T>(PhantomData<T>);

#[cfg(feature = "uuid")]
impl ByteRepresentation for Raw<uuid::Uuid> {
    type Value = uuid::Uuid;
    type DeserializationError = std::array::TryFromSliceError;
    type SerializationError = std::convert::Infallible;

    fn from_bytes(bytes: &[u8]) -> Result<Self::Value, Self::DeserializationError> {
        bytes.try_into().map(uuid::Uuid::from_bytes)
    }

    fn to_bytes(value: &Self::Value) -> Result<Cow<[u8]>, Self::SerializationError> {
        Ok(Cow::Borrowed(value.as_bytes()))
    }
}

impl ByteRepresentation for Raw<u64> {
    type Value = u64;
    type DeserializationError = std::array::TryFromSliceError;
    type SerializationError = std::convert::Infallible;

    fn from_bytes(bytes: &[u8]) -> Result<Self::Value, Self::DeserializationError> {
        bytes
            .try_into()
            .map(Self::Value::from_be_bytes)
    }

    fn to_bytes(value: &Self::Value) -> Result<Cow<[u8]>, Self::SerializationError> {
        Ok(Cow::Owned(value.to_be_bytes().to_vec()))
    }
}

impl ByteRepresentation for Raw<Vec<u8>> {
    type Value = Vec<u8>;
    type DeserializationError = std::convert::Infallible;
    type SerializationError = std::convert::Infallible;

    fn from_bytes(bytes: &[u8]) -> Result<Self::Value, Self::DeserializationError> {
        Ok(bytes.to_vec())
    }

    fn to_bytes(value: &Self::Value) -> Result<Cow<[u8]>, Self::SerializationError> {
        Ok(Cow::Borrowed(value))
    }
}

impl ByteRepresentation for Raw<String> {
    type Value = String;
    type DeserializationError = std::string::FromUtf8Error;
    type SerializationError = std::convert::Infallible;

    fn from_bytes(bytes: &[u8]) -> Result<Self::Value, Self::DeserializationError> {
        Self::Value::from_utf8(bytes.to_owned())
    }

    fn to_bytes(value: &Self::Value) -> Result<Cow<[u8]>, Self::SerializationError> {
        Ok(Cow::Borrowed(value.as_bytes()))
    }
}
